# -*- coding: utf-8 -*-

# (C) Copyright (C) 2013, 2013 Pablo Angulo
# This file is part of karakolas <karakolas.org>.

# karakolas is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as 
# published by the Free Software Foundation, either version 3 of the 
# License, or (at your option) any later version.

# karakolas is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU Affero General Public 
# License along with karakolas.  If not, see 
# <http://www.gnu.org/licenses/>.

#########################################################################

"""
Our MiniPython grammar, the AST classes contain those arguments as properties.

module MiniPython 
{
    mod = Module(stmt* body)
        | Expression(expr body)

    stmt = Assign(expr* targets, expr value)
          | If(expr test, stmt* body, stmt* orelse)
          | Expr(expr value)

    expr = BoolOp(boolop op, expr* values)
         | BinOp(expr left, operator op, expr right)
         | UnaryOp(unaryop op, expr operand)
         | IfExp(expr test, expr body, expr orelse)
         | Compare(expr left, cmpop* ops, expr* comparators)
         | Call(expr func, expr* args, keyword* keywords, expr? starargs, expr? kwargs)
         | Num(object n)
         | Str(string s)

         | Name(identifier id, expr_context ctx)

    expr_context = Load | Store | Del | AugLoad | AugStore | Param

    boolop = And | Or

    operator = Add | Sub | Mult | Div 

    unaryop = Not | UAdd | USub

    cmpop = Eq | NotEq | Lt | LtE | Gt | GtE | In | NotIn

    arguments = (expr* args, identifier? vararg, identifier? kwarg, expr* defaults)
    keyword = (identifier arg, expr value)

    alias = (identifier name, identifier? asname)
}
"""
import ast
from math import floor
from decimal import Decimal, InvalidOperation
from itertools import chain

PRECISION = Decimal('0.01')

class RuleParsingError(Exception):
    pass

class DataProcessingError(Exception):
    pass

def is_printable(c):
    n = ord(c)
    return 32 <= n <=126

def quantize(x):
    return Decimal(x).quantize(PRECISION)

def col2index(col):
    inds = [ord(c)-64 for c in col]
    return sum(d*26**k-1 for k,d in enumerate(reversed(inds)))

class RuleParser(ast.NodeVisitor):
    builtin = {
        'max':max,
        'min':min,
        'round':round,
        'floor':floor,
        'str':str,
        'float':float,
        'int':int,
        'bool':bool,
        'col':col2index,
    }
    def parser_for_type(self, dtype):
        if dtype == 'boolean':
            return bool
        elif dtype == 'integer':
            return int
        elif dtype == 'string':
            return unicode
        elif dtype == 'float' or dtype.startswith('decimal'):
            return quantize
        return lambda s:s

    def __init__(self, rules, fields, defaults = None):
        '''Puede lanzar SyntaxError
        ''' 
        self.tree    = ast.parse(rules)
        self.fields  = fields
        self.skel    = defaults or {}
        self.context = ''
        #Global instructions
        self.extra_cols = dict((node.args[0].s, node.args[1].s)
                           for node in ast.walk(self.tree)
                           if type(node)==ast.Call and node.func.id=='define_extra_col')
#        self.extra_cols = {}
#        for node in ast.walk(self.tree):
#            if type(node)==ast.Call:
#                if node.func.id=='define_extra_cols':
#                    self.extra_cols[node.args[0].s] = node.args[1].s
#                elif node.func.id=='set_context_col':
#                    self.context_col = col2index(node.args[0].s)
        self.parsers = {}
        for k,v in chain(fields.iteritems(), self.extra_cols.iteritems()):
            if v not in self.parsers:
                self.parsers[v] = self.parser_for_type(v)

    def process(self, sheet, row, data, local_vars=None):
        self.data   = data
        self.locals = local_vars or {}
        self.locals.update( {'True':True,'False':False} )

        try:
            self.visit(self.tree)
            prod_data  = dict(self.skel)
            extra_data = {}
            for k,v in self.locals.iteritems():
                if k in self.fields:
                    parse = self.parsers[self.fields[k]]
                    prod_data[k] = parse(v)
                elif k in self.extra_cols:
                    parse = self.parsers[self.extra_cols[k]]
                    extra_data[k] = parse(v)
            return prod_data, extra_data
        except (TypeError, InvalidOperation, DataProcessingError) as e:
            try:
                self.context = unicode(data[self.context_col])
                return {},{}
            except:
                raise DataProcessingError,e
                    
    def name(self, node):
        return node.__class__.__name__

    def visit(self, node):
        if isinstance(node, list):
            return [self.visit(n) for n in node]
        try:
            visitor = getattr(self, 'visit_' + self.name(node))
        except AttributeError:
            raise RuleParsingError("syntax not supported (%s)" % self.name(node))

        return visitor(node)

    def visit_Module(self, node):
        for stmt in node.body:
            self.visit(stmt)

    def visit_Assign(self, node):
        val = self.visit(node.value)
        target = node.targets[0]
        if isinstance(val, (tuple, list)):
            self.locals.update(zip([elt.id for elt in target.elts],val))
        else:
            self.locals[target.id] = val

    def visit_Expr(self, node):
        self.visit(node.value)

    def visit_Name(self, node):
        try:
            val   = self.locals[node.id]
            dtype = self.fields.get(node.id) or self.extra_cols.get(node.id)
            if dtype:
                try:
                    val = self.parsers[dtype](val)
                except (TypeError, InvalidOperation, ValueError):
                    raise DataProcessingError, T("Variable %s was supposed to be of type %s, but it's not")%(node.id, dtype)
            return val
        except KeyError:
            raise RuleParsingError, T('Variable %s was referenced before it was defined')%node.id

    def visit_Num(self, node):
        return Decimal(node.n)

    def visit_Str(self, node):
        assert all(is_printable(c) for c in node.s)
        return node.s

    def visit_Call(self, node):
        if node.func.id == 'cols':
            return self.data
        elif node.func.id == 'context':
            return self.context
        elif node.func.id == 'guess':
            return 'guess'
        elif node.func.id in self.builtin:
            return self.builtin[node.func.id](*[self.visit(arg) for arg in node.args])
        elif node.func.id=='set_context_col':
            self.context_col = col2index(node.args[0].s)        

    def visit_BoolOp(self, node):
        op = type(node.op)
        if op==ast.Or:
            f = lambda a,b:a or b
        if op==ast.And:
            f = lambda a,b:a and b
        return reduce(f, (self.visit(v) for v in node.values))

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.UAdd):
            return self.visit(node.operand)
        if isinstance(node.op, ast.USub):
            return - self.visit(node.operand)
        if isinstance(node.op, ast.Not):
            return not self.visit(node.operand)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        if isinstance(node.op, ast.Add):
            return left + right
        if isinstance(node.op, ast.Sub):
            return left - right
        if isinstance(node.op, ast.Mult):
            return left * right
        if isinstance(node.op, ast.Div):
            return left / right

    def visit_Compare(self, node):
        assert len(node.ops) == 1
        assert len(node.comparators) == 1
        op = node.ops[0]
        comp = self.visit(node.comparators[0])
        left = self.visit(node.left)
        if isinstance(op, ast.Eq):
            return left == comp
        elif isinstance(op, ast.NotEq):
            return left != comp
        elif isinstance(op, ast.Lt):
            return left < comp
        elif isinstance(op, ast.LtE):
            return left <= comp
        elif isinstance(op, ast.Gt):
            return left > comp
        elif isinstance(op, ast.GtE):
            return left >= comp
        elif isinstance(op, ast.In):
            assert isinstance(comp, (str,unicode))
            assert isinstance(left, (str,unicode))
            return left.lower() in comp.lower()
        elif isinstance(op, ast.NotIn):
            assert isinstance(comp, (str,unicode))
            assert isinstance(left, (str,unicode))
            return left.lower() not in comp.lower()

    def visit_IfExp(self, node):
        if self.visit(node.test):
            return self.visit(node.body)
        else:
            return self.visit(node.orelse)

    def visit_If(self, node):
        if self.visit(node.test):
            self.visit(node.body)
        else:
            self.visit(node.orelse)


def test1():
    s='''#pone nombres a las columnas de la tabla
nombre, variedad, origen, _, precio, unidad, descripcion, IVA = cols()
'''
    data = ['naranja', 'navelina', 'Valencia', '','1.24', 'kg', 'ultimas semanas','3.4']
    defaults = {'categoria':db.categorias_productos(codigo=-1).id,
                'productor':1}
    fields = dict((f.name,f.type) for f in db.producto)
    rp = RuleParser(s, fields, defaults)
    print rp.extra_cols
    print repr( rp.process(0,3,data) )


def test2():
    s='''#pone nombres a las columnas de la tabla
producto, variedad, origen, _, precio, unidad, descripcion, IVA = cols()
algo = 2
nombre = producto + variedad
'''
    data = ['naranja', 'navelina', 'Valencia', '','1.24', 'kg', 'ultimas semanas','3.4']
    defaults = {'categoria':db.categorias_productos(codigo=-1).id,
                'productor':1}
    fields = dict((f.name,f.type) for f in db.producto)
    rp = RuleParser(s, fields, defaults)
    print rp.extra_cols
    print repr( rp.process(0,3,data) )

def test3():
    s='''#pone nombres a las columnas de la tabla
producto, variedad, origen, _, precio, unidad, descripcion, IVA = cols()
algo = 2
nombre = producto + variedad
define_extra_col('IVA','decimal')
define_extra_col('algo','string')
'''
    data = ['naranja', 'navelina', 'Valencia', '','1.24', 'kg', 'ultimas semanas','3.4']
    defaults = {'categoria':db.categorias_productos(codigo=-1).id,
                'productor':1}
    fields = dict((f.name,f.type) for f in db.producto)
    rp = RuleParser(s, fields, defaults)
    print rp.extra_cols
    print repr( rp.process(0,3,data, {'fila':3, 'hoja':0}) )


def importa_plantilla(ipedido):
    '''Puede lanzar SyntaxError
    '''
    import xlrd
    from collections import defaultdict

    pedido     = db.pedido(ipedido)
    iproductor = pedido.productor
    icat_none  = db.categorias_productos(codigo=-1).id

    _, xls    = db.pedido.xls.retrieve(pedido.xls)    
    p = xlrd.open_workbook(xls.name)

    defaults = {'categoria':db.categorias_productos(codigo=-1).id,
                'productor':iproductor,
                'activo':True }
    fields = dict((f.name,f.type) for f in db.producto)
    rp = RuleParser(pedido.reglas, fields, defaults)

    db(db.columna_extra.productor==iproductor).delete()
    db(db.producto.productor==iproductor).update(activo=False)

    extracols_dict = {}
    for w in rp.extra_cols:
        cid = db.columna_extra.update_or_insert(productor=iproductor, nombre=w, tipo='string')
        extracols_dict[w] = cid

    nombres = defaultdict(int)
    for k,s in enumerate(p.sheets()):
        for j in xrange(s.nrows):
            try:
                d_prod, d_extra_cols = rp.process(k,j,[c.value for c in s.row(j)], {'fila':j, 'hoja':k})
            except DataProcessingError:
                continue
            #precio positivo y nombre no vacio
            if not d_prod.get('precio_base') or not d_prod.get('nombre'):
                continue
            #TODO ? no estoy usando el ctype de xlrd (antes lo usaba, pero es nec?)
#            cell_precio = d_cells['precio_base']
#            if (cell_precio.ctype!=xlrd.XL_CELL_NUMBER 
#                or not cell_precio.value 
#                or not d_cells['nombre'].value.strip()) :
#                continue
            #nombre
            nombre = d_prod['nombre'].strip()
            if nombre in nombres:
                d_prod['nombre'] = nombre + '_' + str(nombres[nombre]+1)
            else:
                d_prod['nombre'] = nombre
            #TODO ?
            #procesar otros campos tipo 'guess'
            #for k,v ... if v=='guess':...

            #categoria
            cat = d_prod['categoria']
            if cat=='guess':
                d_prod['categoria'] = bayes.adivina_categorias(nombre)
                if d_prod['categoria'] == icat_none:
                    d_prod['categoria'] = bayes.adivina_categorias(rp.context)
            else:
                #si es un string, buscamos la categoria con ese nombre, si es un codigo...
                try:
                    codigo = int(cat)
                    cat = db.categorias_productos(codigo=codigo)
                except ValueError:
                    cat = db(db.categorias_productos.nombre.contains(d_prod['categoria'].encode('utf8'))).select().first()
                if cat:
                    d_prod['categoria'] = cat.id
                else:
                    d_prod['categoria'] = icat_none
            
            #precio_final
            if 'precio_final' not in d_prod:
                d_prod['precio_final'] = d_prod['precio_base']

            try:
                pkeys = dict(nombre=d_prod['nombre'], productor=iproductor)
                pid = _update_or_insert(db.producto, pkeys, d_prod)
#                pid = db.producto.insert(**d_prod)
                nombres[nombre] += 1
                for col, data in d_extra_cols.iteritems():
                    db.valor_extra_string.insert(columna=extracols_dict[col], valor=data, producto=pid)
            except:
                pass


